/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

/*
 * Created on May 23, 2005
 *
 */
package com.sunlabs.archsim.runner;

import java.lang.reflect.Method;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import com.sunlabs.archsim.base.hierarchy.Entities;
import com.sunlabs.archsim.base.hierarchy.Entity;
import com.sunlabs.archsim.base.hierarchy.CompositeEntity;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.build.GASPModelBuilder;
import com.sunlabs.archsim.build.ModelDescription;
import com.sunlabs.archsim.build.ModelNotBuiltException;
import com.sunlabs.archsim.classRegistry.ClassRegistry;
import com.sunlabs.archsim.engine.Director;
import com.sunlabs.archsim.fleet.parser.program.ProgramParser;
import com.sunlabs.archsim.globals.Globals;

/**
 * @author ib27688
 *
 * Runner executes a list of Operations.
 * Also maintains the context - class registry, model description,
 * GASP model group
 */
public class Runner {
	
	public static final String LOAD_PARAMETERS= "load_parameters";
	public static final String PRINT_PARAMETERS= "print_parameters";
	public static final String LOAD_COMPONENTS= "load_components";
	public static final String LOAD_MODEL= "load_model";
	public static final String BUILD_GASP_MODEL= "build_GASP_model";
	public static final String OUTPUT_FILE= "output";
	public static final String JOURNAL_FILE= "journal";
	public static final String ERROR_FILE= "error";
	public static final String INFO_FILE= "info";
	public static final String RUN= "run";
	public static final String FILE= "file";
	public static final String STEP= "step";
	public static final String RESET= "reset";
	public static final String CLEAR_ALL= "clear_all";
	public static final String SELF_CHECK= "self_check";
	public static final String NUMBER= "number";
	public static final String PRINT_COMPONENTS= "print_components";
	public static final String PRINT_MODEL= "print_model";
	public static final String PRINT_DESCRIPTION= "print_model_description";
	public static final String PRINT_STATUS= "print_status";
	public static final String LOAD_ALIASES= "load_aliases";
	public static final String LOAD_PROGRAM= "load_program";
	
	
	
	// context maintained by this object
	private Director director= Director.getInstance();
	private ClassRegistry registry;
	private ModelDescription modelDescription;
	private CompositeEntity model;
	private GASPModelBuilder builderGASP;
	private Globals globals= Globals.getInstance();
	
	public Runner() {
		registry= ClassRegistry.instance;
		modelDescription= new ModelDescription();
		builderGASP= new GASPModelBuilder();
		model= null;
	} // Runner constructor
	
	
	/** 
	 * Execute a list of operations
	 * @param ops
	 */
	public void perform(List ops) {
		Iterator li= ops.iterator();
		while (li.hasNext()) {
			Operation op= (Operation)li.next();
			if (op.name.equals(OUTPUT_FILE)) {
				director.info(OUTPUT_FILE+" " + op.attrNames[0]+ " = " + op.attrValues[0]);
				director.setOutputFile(op.attrValues[0]);
			}
			else if (op.name.equals(JOURNAL_FILE)) {
				director.setJournalFile(op.attrValues[0]);
				director.info(JOURNAL_FILE+" " + op.attrNames[0]+ " = " + op.attrValues[0]);				
			}
			else if (op.name.equals(ERROR_FILE)) {
				director.setErrorFile(op.attrValues[0]);
				director.info(ERROR_FILE+" " + op.attrNames[0]+ " = " + op.attrValues[0]);
			}
			else if (op.name.equals(INFO_FILE)) {
				director.setInfoFile(op.attrValues[0]);
				director.info(INFO_FILE+" " + op.attrNames[0]+ " = " + op.attrValues[0]);
			}
			else if (op.name.equals(LOAD_COMPONENTS)) {
				director.info(LOAD_COMPONENTS+" " + op.attrNames[0]+ " = " + op.attrValues[0]);				
				loadClassRegistry(op.attrValues[0]);
			}
			else if (op.name.equals(LOAD_MODEL)) {
				director.info(LOAD_MODEL+" " + op.attrNames[0]+ " = " + op.attrValues[0]);
				loadModelDescription(op.attrValues[0]);
			}			
			else if (op.name.equals(LOAD_PARAMETERS)) {
				director.info(LOAD_PARAMETERS+" " + op.attrNames[0]+ " = " + op.attrValues[0]);
				loadParameters(op.attrValues[0]);
				// comes after parameters to use the newly used parameters
				// useful in testing
				director.info(LOAD_PARAMETERS+" " + op.attrNames[0]+ " = " + op.attrValues[0]);

			}
			else if (op.name.equals(PRINT_PARAMETERS)) {
				director.info(PRINT_PARAMETERS);
				director.info(globals.toString());
			}
			else if (op.name.equals(BUILD_GASP_MODEL)) {
				director.info(BUILD_GASP_MODEL);
				try {
					buildGASPModel();
				}
				catch (ModelNotBuiltException e) {
					director.error(e.getMessage());
				}
			}
			else if (op.name.equals(RUN)) {
				director.info(RUN+" " + op.attrNames[0]+ " = " + op.attrValues[0]);

				// director.info("Queue size: " + director.queueSize());
				long start= (new Date()).getTime();
				run(op.attrValues[0]);
		 		long stop= (new Date()).getTime();
		 		long runtime= stop-start;
				System.out.println("Done in " + runtime + "ms.");						
			}
			else if (op.name.equals(STEP)) {
				String tail=  (op.attrNames.length==0)? "" : op.attrNames[0]+ " = " + op.attrValues[0];
				director.info(STEP + " " + tail);
				int steps;
				try {
					if (op.attrValues.length == 0) {
						steps= 0;
					}
					else {
						steps= Integer.parseInt(op.attrValues[0]);		
					}
					
					// director.info("Queue size: " + director.queueSize());
					long start= (new Date()).getTime();
					step(steps);
			 		long stop= (new Date()).getTime();
			 		long runtime= stop-start;
					System.out.println("Done in " + runtime + "ms.");					
					
				}
				catch (Exception e) {
					director.error(e.getMessage());
				}
			}
			else if (op.name.equals(LOAD_ALIASES)) {
				director.info(LOAD_ALIASES);
				ProgramParser parser= new ProgramParser();
				try {
					parser.parse(op.attrValues[0]);		}
				catch (Exception e) {
					director.fatalError("Attribute parsing error: " + e.getMessage());
				}
			}
			else if (op.name.equals(LOAD_PROGRAM)) {
				director.info(LOAD_PROGRAM);
				ProgramParser parser= new ProgramParser();
				try {
					parser.parse(op.attrValues[0]);		}
				catch (Exception e) {
					director.fatalError("Program parsing error: " + e.getMessage());
				}
			}
			else if (op.name.equals(RESET)) {
				director.info(RESET);
				reset();
			}
			else if (op.name.equals(CLEAR_ALL)) {
				director.info(CLEAR_ALL);
				clearAll();
			}
			else if (op.name.equals(SELF_CHECK)) {
				director.info(SELF_CHECK);
				selfCheck();
			}
			else if (op.name.equals(PRINT_COMPONENTS)) {
				director.info(PRINT_COMPONENTS);
				printComponents();
			}
			else if (op.name.equals(PRINT_MODEL)) {
				director.info(PRINT_MODEL);
				model.print();
			}
			else if (op.name.equals(PRINT_DESCRIPTION)) {
				director.info(PRINT_DESCRIPTION);
				modelDescription.print();
			}
			else if (op.name.equals(PRINT_STATUS)) {
				director.info(PRINT_STATUS);
				Iterator<Entity> ei= Entities.iterator();
				while (ei.hasNext() ) {
					Entity e= ei.next();
					Method m= null;
					try {
						m= e.getClass().getMethod("stateInfo", null);
					}
					catch (SecurityException se) {
						director.fatalError("Runner caught security exception when inspecting class "
								+ e.getClass().getName());
					}
					catch (NoSuchMethodException nme) {
						m= null;
					}
					if (m != null) {
						try {
							// get status
							Object so= m.invoke(e, null );
							// print status
							director.info(so.toString());
						}
						catch (Exception ie) {
							director.fatalError("Runner caught an exception when invoking "
									+ "method "
									+ m.getName() 
									+ " on object" + e
									+ ". Exception message= "
									+ ie.getMessage());
						}
						
					}
				}
			}
			else {
				// simply ignore an unknown operation
				// TODO - consider other options
				director.error("Warning: ignoring unknown operation: " +
						op.name);
			}
		}
		
	} // perform
	
	public ClassRegistry loadClassRegistry(String file) {
		registry.load(file);
		return registry;
	} // loadClassRegistry

	public Globals loadParameters(String file) {
		globals.load(file);
		return globals;
	} // loadClassRegistry
	
	public ModelDescription loadModelDescription(String file) {
		if (registry == null) return null;
		modelDescription.load(file);
	    return modelDescription;
	} // loadModel
	
	public CompositeEntity buildGASPModel()
	throws ModelNotBuiltException {
		model= builderGASP.buildModel(modelDescription);
		return model;
	} // buildModel
	
	/**  
	 * Run for the specified time.
	 * @param t tume
	 */
	public void run(String t) {
		try {
			// convert the time to integer
			int time= Integer.parseInt(t);
			director.run(new Delay(time));
		}
		catch (Exception e){
			e.printStackTrace();
			director.error("Error: " + e.getMessage());
		}
	} // run
	/**
	 * Run for the specified delay
	 * @param t delay
	 */
	public void run(int t) {
		director.run(new Delay(t));
	} // run

	/**
	 * Step the simulation for the specified 
	 * number of steps.
	 * @param steps the number of steps
	 */
	public void step(String steps) {
		try {
			int s= Integer.parseInt(steps);
			director.step(s);
		}
		catch (Exception e) {
			director.error("Error: " + e.getMessage());
		}
	} // step
	
	/** 
	 * Step the simulation for the specified 
	 * number of steps.
	 * @param steps the number of steps
	 */
	public void step(int steps) {
		for (int i=0; i<steps; i++) 
			director.step();
	}
	
	/**
	 * Reset the simulation.
	 */
	public void reset() {
		director.reset();
	} // reset

	/**
	 * Reset the context and the simulation.
	 */
	public void clearAll() {
		ClassRegistry.getInstance().clear();
		modelDescription= new ModelDescription();
		director.reset();
		model= null;
		globals.clear();
	} // clearAll

	/**
	 * Self check for the model.
	 */
	public void selfCheck() {
		if (model != null ) {
			director.info("Self-check: " + model.selfCheck());
		}
		else {
			director.error("Self check requested on null model");
		}
			
	} // Self check

	
	public void printComponents() {
		director.info(ClassRegistry.getInstance().toString());
	} // printComponents
	
	public void printModel() {
		modelDescription.print();
	} // printModel
	
	public void printModelDescription() {
		model.print();
	} // printModelDescription
	
	public static void main(String[] args) {
		
		Director director= Director.getInstance();
		
		if (args.length != 1) {
			director.error("Exactly one argument required");
			director.closeJournal();
			System.exit(1);
		}
		String fileName= args[0];
		OperationLoader ldr= new OperationLoader();
		List ops= ldr.load(fileName);
		
		Runner r= new Runner();
		r.perform(ops);
		
		director.closeJournal();
		
	} // main
	
} // Runner
